local _G = _G
local Obituary = _G.Obituary
local L = Obituary.L

local GUID = LibStub("LibGUIDRegistry-0.1")

local bit_bor, bit_band = _G.bit.bor, _G.bit.band
local AFFILIATION_IN_GROUP = bit_bor(COMBATLOG_OBJECT_AFFILIATION_PARTY, COMBATLOG_OBJECT_AFFILIATION_RAID, COMBATLOG_OBJECT_AFFILIATION_MINE)
local format, strsub, gsub, strlen = _G.string.format, _G.string.sub, _G.string.gsub, _G.string.len

-- This should actually be 255, but I'm keeping it at 250 for the sake of laziness in length checking in the reporting functions.  Yay lazy!
local MAX_MESSAGE_LENGTH = 250

local defaultOutputChannels = Obituary.defaultOutputChannels

local deathEventHistory, deathReport = {}, {}
local ChatFrameList = {}
do
	for i = 1, NUM_CHAT_WINDOWS do
		ChatFrameList[i] = getglobal("ChatFrame" .. i)
	end
end

function Obituary:PurgeDeathEventHistories()
	for k, v in pairs(deathEventHistory) do
		deathEventHistory[k] = nil
	end
end

function Obituary:PurgeDeathReports()
	for k, v in pairs(deathReport) do
		deathReport[k] = nil
	end
end

function Obituary:COMBAT_LOG_EVENT_UNFILTERED(_, timestamp, event, hideCaster, srcGUID, srcName, srcFlags, srcRaidFlags, dstGUID, dstName, dstFlags, dstRaidFlags, ...)
	if bit_band(dstFlags, COMBATLOG_OBJECT_TYPE_PLAYER) == 0 then
		if bit_band(dstFlags, COMBATLOG_OBJECT_TYPE_PET) ~= 0 or bit_band(dstFlags, COMBATLOG_OBJECT_TYPE_GUARDIAN) ~= 0 then
			-- This is an event for either a pet or a guardian.  If we're not interested in this type of event, ignore it.
			if not Obituary.db.EnableForUnits.Pets then return end
		else
			-- This is a non-pet/guardian event, so we're going to assume it's just a general NPC event.  If we're not interested in this type of event, ignore it.
			if not Obituary.db.EnableForUnits.NPCs then return end
		end
	else
		if bit_band(dstFlags, AFFILIATION_IN_GROUP) ~= 0 then
			-- This is an event for either yourself, or someone that you're grouped with.  If we're not interested in this type of event, ignore it.
			if not Obituary.db.EnableForUnits.PlayersInGroup then return end
		else
			-- This is an event for someone outside your group.  If we're not interested in this type of event, ignore it.
			if not Obituary.db.EnableForUnits.PlayersNotInGroup then return end
		end
	end

	local abilityID, abilityName, amount, amountOver, resisted, blocked, absorbed, isCrit, isCrush, isHeal = nil, nil, nil, 0, 0, 0, 0, false, false, false

	local prefix, suffix, extra = strsplit("_", event)

	if event == "UNIT_DIED" then
		if UnitIsFeignDeath(dstName) then return end
		self:ReportDeath(timestamp, dstGUID, dstName, dstFlags)
	elseif prefix == "SPELL" then
		if suffix == "DAMAGE" or extra == "DAMAGE" then		-- SPELL_DAMAGE, SPELL_PERIODIC_DAMAGE
			abilityID, abilityName = select(1, ...)
			amount, amountOver = select(4, ...)
			resisted, blocked, absorbed, isCrit = select(7, ...)
			isCrush = select(12, ...)
		elseif suffix == "HEAL" or extra == "HEAL" then		-- SPELL_HEAL, SPELL_PERIODIC_HEAL
			abilityID, abilityName = select(1, ...)
			amount, amountOver, isCrit = select(4, ...)
			isHeal = true
		elseif suffix == "INSTAKILL" then			-- SPELL_INSTAKILL
			abilityID, abilityName = select(1, ...)
			amount = -1
		end
	elseif suffix == "DAMAGE" then
		if prefix == "SWING" then				-- SWING_DAMAGE
			abilityName = L["melee swing"]
			amount, amountOver = select(1, ...)
			resisted, blocked, absorbed, isCrit = select(4, ...)
			isCrush = select(9, ...)
		elseif prefix == "RANGE" then				-- RANGE_DAMAGE
			abilityID, abilityName = select(1, ...)
			amount, amountOver = select(4, ...)
			resisted, blocked, absorbed, isCrit = select(7, ...)
			isCrush = select(12, ...)
		elseif prefix == "ENVIRONMENTAL" then			-- ENVIRONMENTAL_DAMAGE
			abilityName, amount, amountOver = select(1, ...)
			resisted, blocked, absorbed, isCrit = select(5, ...)
			isCrush = select(10, ...)
		end
	elseif prefix == "DAMAGE" and extra ~= "MISSED" then		-- DAMAGE_SHIELD but not DAMAGE_SHIELD_MISSED
		abilityID, abilityName = select(1, ...)
		amount, amountOver = select(4, ...)
		resisted, blocked, absorbed, isCrit = select(7, ...)
		isCrush = select(12, ...)
	end

	if amount then
		self:RecordEvent(timestamp, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, abilityID, abilityName, amount, amountOver, resisted, blocked, absorbed, isCrit, isCrush, isHeal)
	end
end

local raidTargetText = {"{star}", "{circle}", "{diamond}", "{triangle}", "{moon}", "{square}", "{cross}", "{skull}"}
local raidTargetPlainText = {"ST ", "CR ", "DI ", "TR ", "MN ", "SQ ", "X ", "SK "}
local raidTargetColoredText = {"|cfffffd61ST|r ", "|cffffb361CR|r ", "|cffd861ffDI|r ", "|cff7fff61TR|r ", "|cffb0f8ffMN|r ", "|cff2d9efeSQ|r ", "|cfffe2d2dX|r ", "|cffc8c8c8SK|r "}
function Obituary:GetRaidTargetIcon(index, useIcon)
	if index > 0 and index <= 8 then
		if not useIcon then
			if Obituary.db.Reporting.UseGraphicalIcons then
				return raidTargetText[index]
			else
				return raidTargetPlainText[index]
			end
		else
			if Obituary.db.General.UseGraphicalIcons then
				local path = COMBATLOG_OBJECT_RAIDTARGET1 * (index ^ (index - 1))
				return "|Hicon:" .. path .. ":dest|h|TInterface\\TargetingFrame\\UI-RaidTargetingIcon_" .. index .. ".blp:0|t|h"
			else
				return raidTargetColoredText[index]
			end
		end
	end

	return nil
end

function Obituary:GetRaidTargetIconByFlags(flags, useIcon)
	if bit_band(flags, COMBATLOG_OBJECT_SPECIAL_MASK) ~= 0 then
		for index = 1, 8 do
			local bitMask = COMBATLOG_OBJECT_RAIDTARGET1 * (2 ^ (index - 1))
			local raidIcon = (bit_band(flags, bitMask) == bitMask)
			if raidIcon then
				return self:GetRaidTargetIcon(index, useIcon)
			end
		end
	end

	return nil
end

function Obituary:GetColoredName(guid, name, flags)
	if bit_band(flags, COMBATLOG_OBJECT_TYPE_PLAYER) == 0 then
		-- This isn't a player, so lets use the reaction color.
		if bit_band(flags, COMBATLOG_OBJECT_REACTION_HOSTILE) ~= 0 then
			-- The NPC in question is hostile, so color it red.
			return "|cffff0000" .. name .. "|r"
		elseif bit_band(flags, COMBATLOG_OBJECT_REACTION_FRIENDLY) ~= 0 then
			-- The NPC in question is friendly, so color it green.
			return "|cff00ff00" .. name .. "|r"
		elseif bit_band(flags, COMBATLOG_OBJECT_REACTION_NEUTRAL) ~= 0 then
			-- The NPC in question is neutral, so color it yellow.
			return "|cffffff00" .. name .. "|r"
		end

		-- Default to white.
		return "|cffffffff" .. name .. "|r"
	else
		-- This is a player, so lets use the default class color if we can determine their class.
		local classColor = nil
		-- First, try to get the class via LibGUIDRegistry.  If that doesn't work, fall back to UnitClass.
		local class = GUID:GetClass(guid)
		if class then
			classColor = RAID_CLASS_COLORS[class]
		elseif UnitIsPlayer(name) then
			_, class = UnitClass(name)
			classColor = RAID_CLASS_COLORS[class]
		end

		if classColor then
			return format("|cff%02x%02x%02x%s|r", classColor.r * 255, classColor.g * 255, classColor.b * 255, name)
		end

		-- Default to grey.
		return "|cff888888" .. name .. "|r"
	end
end

function Obituary:DecorateName(guid, name, flags, localOutput)
	if not name then return nil end

	if localOutput then
		name = self:GetColoredName(guid, name, flags)
	end
	local raidTargetIcon = self:GetRaidTargetIconByFlags(flags, localOutput)

	if raidTargetIcon then
		return raidTargetIcon .. name
	end
	return name
end

function Obituary:RecordEvent(timestamp, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, abilityID, abilityName, amount, amountOver, resisted, blocked, absorbed, isCrit, isCrush, isHeal)
	-- If UnitHealthMax returns non-zero, we can provide current and maximum health values for this unit.
	local curHealth, maxHealth = nil, nil
	if UnitHealthMax(dstName) ~= 0 then
		curHealth = UnitHealth(dstName)
		maxHealth = UnitHealthMax(dstName)
	end

	-- Ensure that the table is properly initialized.
	if not deathEventHistory[dstGUID] then
		deathEventHistory[dstGUID] = Stack.new(math.max(Obituary.db.Reporting.Summary.ReportEvents, Obituary.db.Reporting.Detailed.ReportEvents))
	end

	-- Sanitize and store the event information.  Apparently values that should be zero can also be sent to us as nil instead.
	Stack.push(deathEventHistory[dstGUID], {
		timestamp = timestamp,
		srcGUID = srcGUID,
		srcName = srcName,
		srcFlags = srcFlags,
		abilityID = abilityID,
		abilityName = abilityName,
		amount = amount,
		amountOver = amountOver or 0,
		resisted = resisted or 0,
		blocked = blocked or 0,
		absorbed = absorbed or 0,
		isCrit = isCrit,
		isCrush = isCrush,
		isHeal = isHeal,
		curHealth = curHealth,
		maxHealth = maxHealth,
	})
end

function Obituary:ReportDeath(timestamp, guid, name, flags)
	-- If we don't have any events stored, then there's nothing to report.
	if not deathEventHistory[guid] or Stack.isEmpty(deathEventHistory[guid]) then return end

	-- Are we supposed to print a detailed death report?
	if Obituary.db.Reporting.Detailed.Display then
		-- Ensure that deathReport[guid] is initialized.
		if not deathReport[guid] then
			deathReport[guid] = {}
		end
		local currentDeath = tostring(timestamp)
		if not deathReport[guid][currentDeath] then
			deathReport[guid][currentDeath] = Stack.new(Obituary.db.Reporting.Detailed.ReportEvents)
		end

		-- Iterate through the event history, storing the events that haven't expired.
		local lastDamageSource = nil
		Stack.iterator(deathEventHistory[guid])
		local event = Stack.next(deathEventHistory[guid])
		while event do
			if timestamp - event["timestamp"] <= Obituary.db.General.EventTTL then
				Stack.push(deathReport[guid][currentDeath], event)
				if not event["isHeal"] then
					lastDamageSource = event["srcName"] or L["DefaultMobName"]
				end
			end
			event = Stack.next(deathEventHistory[guid])
		end

		-- If we have a last damage source, then we have a valid report.
		if lastDamageSource then
			-- Format the chat link and dump it to the desired chat frame.
			local frame = ChatFrameList[Obituary.db.Reporting.Detailed.ChatFrame]
			if frame then
				frame:AddMessage(format(L["ObituaryChatLink"], timestamp, guid, name, flags, name, lastDamageSource))
			end
		end
	end

	-- Are we supposed to print a summarized death report?
	if Obituary.db.Reporting.Summary.Display then
		self:SummarizeDeathReport(timestamp, guid, name, flags)
	end

	-- Wipe the event history for this unit since we no longer need it.
	deathEventHistory[guid] = nil
end

function Obituary:SummarizeDeathReport(timestamp, guid, name, flags)
	-- If we don't have any events stored, then there's nothing to report.
	if not deathEventHistory[guid] or Stack.isEmpty(deathEventHistory[guid]) then return end

	-- We're only interested in damage events that haven't expired yet.
	local damageEvents = Stack.new(Obituary.db.Reporting.Summary.ReportEvents)
	Stack.iterator(deathEventHistory[guid])
	local event = Stack.next(deathEventHistory[guid])
	while event do
		if timestamp - event["timestamp"] <= Obituary.db.General.EventTTL and not event["isHeal"] then
			Stack.push(damageEvents, event)
		end
		event = Stack.next(deathEventHistory[guid])
	end

	-- We don't have any damage events to report.
	if Stack.isEmpty(damageEvents) then return end

	-- Set some variables that we will be using to generate the report.
	local outputChannel = Obituary.db.Reporting.Summary.Output
	local localOutput = outputChannel == "ChatFrame"
	if localOutput then
		outputChannel = ChatFrameList[Obituary.db.Reporting.Summary.ChatFrame]
	end
	local reportHeader = format(L["SummaryHeader"], self:DecorateName(guid, name, flags, localOutput))
	local lastDamageSource = ""	-- Formatted name of the last source of damage.
	local reportString = ""		-- The report to be displayed.
	local sourceString = ""		-- Temporary string used to store per-target report information.

	-- Iterate through the event history, generating the report.
	while not Stack.isEmpty(damageEvents) do
		event = Stack.pop(damageEvents)

		local srcName = self:DecorateName(event["srcGUID"], event["srcName"] or L["DefaultMobName"], event["srcFlags"], localOutput)
		if lastDamageSource ~= srcName then
			-- This check is just here so that this block isn't executed for the very first event.
			if reportString ~= "" then
				reportString = format("%s), ", strsub(reportString, 1, -3))
			end
			lastDamageSource = srcName
			srcName = nil
			sourceString = format("%s (", lastDamageSource)
		end

		-- Don't show an amount for instakill abilities, such as Divine Intervention.
		local amount = nil
		if event["amount"] ~= -1 then
			amount = format(" %s %d", L["for"], event["amount"])
		end

		-- Format this event.
		sourceString = format("%s%s%s%s%s, ",
			sourceString,
			event["abilityID"] and GetSpellLink(event["abilityID"]) or event["abilityName"],
			amount or "",
			event["isCrit"] and L["crit"] or "",
			event["isCrush"] and L["crush"] or ""
		)

		-- If the sum length of the report header, last damage source, report string, and source string exceed the max message length, send the current report.
		if not localOutput and strlen(reportHeader) + strlen(lastDamageSource) + strlen(reportString) + strlen(sourceString) >= MAX_MESSAGE_LENGTH then
			if srcName then
				reportString = format("%s%s)", reportHeader, strsub(reportString, 1, -3))
				sourceString = format("%s (%s", lastDamageSource, sourceString)
			else
				reportString = format("%s%s", reportHeader, strsub(reportString, 1, -3))
			end
			self:SendMessage(reportString, outputChannel)
			reportString = ""
		end

		reportString = reportString .. sourceString
		sourceString = ""
	end
	reportString = format("%s%s)", reportHeader, strsub(reportString, 1, -3))
	self:SendMessage(reportString, outputChannel)
end

function Obituary:FormatDeathEvent(event, timeOfDeath, localMessage)
	local eventString = ""
	if event["curHealth"] and event["maxHealth"] then
		-- TODO: This seems too simplistic.  Is there some way to get better information here?
		eventString = format("(%d/%d) ", event["maxHealth"], event["curHealth"])
	end
	if event["srcName"] then
		eventString = format("%s%s's ", eventString, self:DecorateName(event["srcGUID"], event["srcName"], event["srcFlags"], localMessage))
	end
	eventString = format("%s%s", eventString, event["abilityID"] and GetSpellLink(event["abilityID"]) or event["abilityName"])
	if event["amount"] ~= -1 then
		local amount = event["amount"]
		if localMessage then
			if event["isHeal"] then
				amount = format(L["ColorHeal"], amount)
			else
				amount = format(L["ColorDamage"], amount)
			end
		end
		eventString = format("%s %s %s", eventString, L["for"], amount)
	end
	if event["amountOver"] > 0 then
		eventString = format("%s (%d %s)", eventString, event["amountOver"], event["isHeal"] and L["overheal"] or L["overkill"])
	end
	if event["isCrit"] then
		eventString = format("%s%s", eventString, L["crit"])
	end
	if event["isCrush"] then
		eventString = format("%s%s", eventString, L["crush"])
	end
	if event["resisted"] > 0 then
		eventString = format("%s (%d %s)", eventString, event["resisted"], L["resisted"])
	end
	if event["blocked"] > 0 then
		eventString = format("%s (%d %s)", eventString, event["blocked"], L["blocked"])
	end
	if event["absorbed"] > 0 then
		eventString = format("%s (%d %s)", eventString, event["absorbed"], L["absorbed"])
	end

	-- FIXME: There's a strange bug here.  Sometimes, timeOfDeath - timestamp will return a negative number.  I hope I've fixed the bug, but I'm not 100% certain that I have.
	local timeBeforeDeath = format("-%.02fs", timeOfDeath - event["timestamp"])

	return eventString, timeBeforeDeath
end

-- TODO: This function needs to somehow handle message splitting based on message length.
function Obituary:SendMessage(msg, channel)
	if type(channel) == "table" then
		-- The desired channel is a chat frame.

		channel:AddMessage(msg)
	elseif defaultOutputChannels[channel] then
		-- The desired channel is a default channel such as "RAID", "GUILD", etc.

		-- We can't report to guild or officer chat if we're not in a guild, so silently fail.
		if (channel == "GUILD" or channel == "OFFICER") and not IsInGuild() then return end
		-- If we're trying to report to raid but UnitInRaid says we're not in a raid, silently fall back to party chat.
		if channel == "RAID" and not UnitInRaid("player") then
			channel = "PARTY"
		end
		-- We can't report to party chat if we're not in a party, so silently fail.
		if channel == "PARTY" and GetNumPartyMembers() == 0 then return end
		SendChatMessage(msg, channel)
	else
		-- The desired channel is a custom channel such as "1", "2", etc.

		-- This was shamelessly stolen from Prat.
		if Obituary.db.Reporting.UseCLINK then
			msg = gsub(msg, "|c(.-)|H(.-):(.-)|h.(.-).|h|r", function(a, b, c, d) return format("{CLINK:%s:%s:%s:%s}", b, a, c, d) end)
		end
		SendChatMessage(msg, "CHANNEL", nil, channel)
	end
end

function Obituary:ParseLink(link, text, button)
	local linkType, linkTimestamp, linkGUID, linkName, linkFlags = strsplit(":", link)
	if linkType == "obit" then
		-- If we don't have information stored for this death, inform the user.
		if not deathReport[linkGUID] or not deathReport[linkGUID][linkTimestamp] or Stack.isEmpty(deathReport[linkGUID][linkTimestamp]) then
			ShowUIPanel(ItemRefTooltip)
			if not ItemRefTooltip:IsShown() then
				ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
			end
			ItemRefTooltip:ClearLines()
			ItemRefTooltip:AddLine(format(L["DeathReportFor"], linkName))
			ItemRefTooltip:AddLine("---")
			ItemRefTooltip:AddLine(L["There appears to be no information stored for this death."])
			return
		end

		if IsShiftKeyDown() then
			-- The Shift key was being pressed while clicking the link.  Output the death report.
			local outputChannel = Obituary.db.Reporting.Detailed.Output
			if outputChannel == "ChatFrame" then
				outputChannel = ChatFrameList[Obituary.db.Reporting.Detailed.ChatFrame]
			end
			self:SendMessage(format(L["DeathReportFor"], self:DecorateName(linkGUID, linkName, linkFlags, false)), outputChannel)

			-- Create the bulk of the report information.
			Stack.iterator(deathReport[linkGUID][linkTimestamp])
			local event = Stack.next(deathReport[linkGUID][linkTimestamp])
			while event do
				local eventString, timeBeforeDeath = self:FormatDeathEvent(event, linkTimestamp, false)
				if eventString then
					eventString = format("%s: %s", timeBeforeDeath, eventString)
					self:SendMessage(eventString, outputChannel)
				end
				event = Stack.next(deathReport[linkGUID][linkTimestamp])
			end
		else
			-- The Shift key was not being pressed while clicking the link.  Display the death report in a tooltip.
			-- TODO: Ideally clicking a link that's already open should close the link.  Need to figure out how to determine if the link that we're clicking is the same one that's being displayed, however.
			ShowUIPanel(ItemRefTooltip)
			if not ItemRefTooltip:IsShown() then
				ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
			end
			ItemRefTooltip:ClearLines()
			ItemRefTooltip:AddLine(format(L["DeathReportFor"], self:DecorateName(linkGUID, linkName, linkFlags, true)))
			ItemRefTooltip:AddLine("---")

			-- Create the bulk of the report information.
			Stack.iterator(deathReport[linkGUID][linkTimestamp])
			local event = Stack.next(deathReport[linkGUID][linkTimestamp])
			while event do
				local eventString, timeBeforeDeath = self:FormatDeathEvent(event, linkTimestamp, true)
				if eventString then
					ItemRefTooltip:AddDoubleLine(eventString, timeBeforeDeath)
				end
				event = Stack.next(deathReport[linkGUID][linkTimestamp])
			end

			ItemRefTooltip:Show()
		end
	else
		return self.hooks["SetItemRef"](link, text, button)
	end
end
